/*
    Copyright  1995-2001, The AROS Development Team. All rights reserved.
    $Id: rexx.c,v 1.2 2005/09/23 10:51:30 piru Exp $

    Desc:
    Lang: English
*/

#include <proto/dos.h>
#include <proto/utility.h>
#include <proto/datatypes.h>
#include <proto/rexxsyslib.h>
#include <rexx/errors.h>
#include "datatypes_intern.h"

#include <string.h>
#include <stdlib.h>
#include <stddef.h>


static void UpdateArgstring(STRPTR string, UWORD length)
{
	struct RexxArg *argstr = (struct RexxArg *)(string - offsetof(struct RexxArg, ra_Buff));
	UBYTE hash = 0;

	argstr->ra_Buff[length] = 0;
	argstr->ra_Length = length;

	while (length--)
	{
		hash += *string++;
	}

	argstr->ra_Hash = hash;
}

static LONG rxExamineDT(struct DataTypesBase *DataTypesBase, struct RexxMsg *rxmsg, STRPTR *resstr)
{
	return -ERR10_001;
}

static LONG rxFindTool(struct DataTypesBase *DataTypesBase, struct RexxMsg *rxmsg, STRPTR *resstr)
{
	return -ERR10_001;
}

static LONG rxGetDTString(struct DataTypesBase *DataTypesBase, struct RexxMsg *rxmsg, STRPTR *resstr)
{
	STRPTR str;
	ULONG id = strtol(rxmsg->rm_Args[1], (char **) &str, 10);

	if (str == rxmsg->rm_Args[1] && LengthArgstring(rxmsg->rm_Args[1]) == 4) id = *((ULONG *)rxmsg->rm_Args[1]);

	if ((str = GetDTString(id)))
	{
		STRPTR tmp, tmp1 = "%s", tmp2;
		int len, len1, len2;

		len = strlen(str);

		if ((tmp2 = strstr(str, tmp1)))
		{
			len2 = tmp2 - str;
			len -= 2;
		}
		else len2 = 2;

		if (((rxmsg->rm_Action & RXARGMASK) == 2) && rxmsg->rm_Args[2])
		{
			tmp1 = rxmsg->rm_Args[2];
			len1 = strlen(tmp1);
		}
		else len1 = 2;

		tmp = CreateArgstring(NULL, len + len1 + len2);

		if (tmp2)
		{
			bcopy(str, tmp, len2);
			str = tmp + len2;

			if (len2 < len) bcopy(tmp2 + 2, str + len1, len - len2);

			bcopy(tmp1, str, len1);
		} else {
			bcopy(tmp1, tmp, len1);
			tmp[len1] = ':';
			tmp[len1 + 1] = ' ';

			bcopy(str, tmp + len1 + len2, len);
		}

		UpdateArgstring((str = tmp), len + len1 + len2);
	} else str = CreateArgstring(NULL, 0);

	*resstr = str;
	return 0;
}

static LONG rxLaunchTool(struct DataTypesBase *DataTypesBase, struct RexxMsg *rxmsg, STRPTR *resstr)
{
	return -ERR10_001;
}

static LONG rxObtainDatatype(struct DataTypesBase *DataTypesBase, struct RexxMsg *rxmsg, STRPTR *resstr)
{
	return -ERR10_001;
}

struct dispentry
{
	CONST_STRPTR FuncName;
	LONG (*Function)(struct DataTypesBase *, struct RexxMsg *, STRPTR *);
	UWORD MinArgs, MaxArgs;
};

/* MUST be alphabetically sorted! */
struct dispentry disptable[] = {
//	{ "DEBUG", rxDebug, 0, 0 },
	{ "EXAMINEDT", rxExamineDT, 1, 3 },
	{ "FINDTOOL", rxFindTool, 1, 5 },
	{ "GETDTSTRING", rxGetDTString, 1, 2 },
	{ "LAUNCHTOOL", rxLaunchTool, 1, 4 },
	{ "OBTAINDATATYPE", rxObtainDatatype, 3, 5 },
};

int dispcomp(const void *name, const void *dispentry)
{
	return strcmp(name, ((const struct dispentry *)dispentry)->FuncName);
}


AROS_LH1(ULONG, RLDispatch,
	 AROS_UFHA(struct RexxMsg *, rxmsg, A0),
        struct DataTypesBase *, DataTypesBase, 5, DataTypes)
{
    AROS_LIBFUNC_INIT

	struct dispentry *dispentry;
	LONG numargs, reslen;
	UBYTE resbuf[12];
	STRPTR argstr = resbuf;

	if (!RexxSysBase) return(ERR10_014);
	if (!rxmsg || !IsRexxMsg(rxmsg) || !(rxmsg->rm_Action & RXFUNC) || !rxmsg->rm_Args[0]) return(ERR10_010);

	if (!(dispentry = bsearch(rxmsg->rm_Args[0], disptable, sizeof(disptable)/sizeof(struct dispentry), sizeof(struct dispentry), dispcomp))) return(ERR10_001);

	numargs = (rxmsg->rm_Action & RXARGMASK);
	if (numargs < dispentry->MinArgs || numargs > dispentry->MaxArgs) return ERR10_017;

	for (reslen=1; reslen<=dispentry->MinArgs; reslen++) if (!rxmsg->rm_Args[reslen]) return ERR10_018;

	reslen = dispentry->Function(DataTypesBase, rxmsg, &argstr);

	if (reslen < 0) return (ULONG)(-reslen);

	if (reslen > 0)
	{
		if (!(argstr = CreateArgstring(argstr, reslen))) return ERR10_003;
	}

#ifdef __MORPHOS__
	REG_A0 = (ULONG)argstr;
#else
#warning "Rexx expects the result in A0...yeah..what a braindead mess"
#endif

    return 0;

    AROS_LIBFUNC_EXIT
}
